// Aqsis
// Copyright (c) 1997 - 2001, Paul C. Gregory
//
// Contact: pgregory@aqsis.com
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/**
 * Copyright (C) 2006-2007  NTT DATA CORPORATION
 * 
 * Version: 1.0.0 2007/04/01
 *  
 */
package net.cellcomputing.himawari.shadervm;

import static net.cellcomputing.himawari.library.EqIntIndex.SHD_Var_varying_init;
import static net.cellcomputing.himawari.library.EqIntIndex.SHD_var_varying_point;
import net.cellcomputing.himawari.accessory.STLArray;
import net.cellcomputing.himawari.accessory.STLVector;
import net.cellcomputing.himawari.accessory.primitive.p_boolean;
import net.cellcomputing.himawari.library.CqStats;
import net.cellcomputing.himawari.library.EqVariableType;
import net.cellcomputing.himawari.library.IqShaderData;
import net.cellcomputing.himawari.library.types.CqVector3D;

/**
 * CqShaderVariableVaryingNXCqVector3Dɓ
 * NX
 * @author NTT DATA Corporation
 */
public strictfp class CqShaderVariableVaryingPoint extends CqShaderVariableVarying<CqVector3D> {
	
	/**
	 * RXgN^B
	 * Vector̔z1ɃZbgāA
	 * vXVB
	 */
	public CqShaderVariableVaryingPoint( )
	{
		super(new EqVariableType(EqVariableType.type_point));
		//Ϗ̂
		//m_aValue = new STLVector<CqVector3D>(CqVector3D.class);
		m_aValue = new STLArray<CqVector3D>(CqVector3D.class);
		m_aValue.resize( 1 );
		CqStats.STATS_INC( SHD_var_varying_point );
	}
	
	/**
	 * RXgN^B
	 * Vector̔z1ɃZbgāA
	 * vXVB
	 * 
	 * @param strName
	 * @param fParameter
	 */
	public CqShaderVariableVaryingPoint( final String strName, boolean fParameter)
	{
		super(strName,fParameter,new EqVariableType(EqVariableType.type_point));
		//Ϗ̂
		//m_aValue = new STLVector<CqVector3D>(CqVector3D.class);
		m_aValue = new STLArray<CqVector3D>(CqVector3D.class);
		m_aValue.resize(1);
		CqStats.STATS_INC( SHD_var_varying_point );
	}
	
	/**
	 * RXgN^B
	 * Vector̔z1ɃZbgāA
	 * vXVB
	 * 
	 * @param strName
	 */
	public CqShaderVariableVaryingPoint( final String strName){
		this(strName,false);
	}
	
	/**
	 * Rs[RXgN^
	 * @param val
	 */
	public CqShaderVariableVaryingPoint(CqShaderVariableVaryingPoint val){
		super(val,new EqVariableType(EqVariableType.type_point));
		//m_aValue = new STLVector<CqVector3D>(CqVector3D.class);
		m_aValue = new STLArray<CqVector3D>(CqVector3D.class);
		m_aValue.resize( val.m_aValue.size() );
//		m_aValue.assign( val.m_aValue.begin(), val.m_aValue.begin() );
		//R[hɏC 2005/11/2 nttdata
		for(int i = 0; i < val.m_aValue.size(); i++)
			m_aValue.elementAt(i).assignment(val.m_aValue.elementAt(i));
	}
	
	/**
	 * u=vZq̃I[o[[h
	 *  Vector̒lRs[B
	 * @param From
	 */
	public void assignment(CqShaderVariableVarying From) {
		m_aValue.resize(From.m_aValue.size());
		for(int i = 0; i < From.m_aValue.size();i++){
			m_aValue.elementAt(i).assignment((CqVector3D)From.m_aValue.elementAt(i));
		}
		
	}
	
	/**
	 * eNXŒ`ꂽ֐
	 * Vector̃TCY1ȏł
	 * w肳ꂽ擪̒lVector𖄂ߐsB
	 * Vector̃TCY0ȉł΁A
	 * lŖߐsB
	 * 
	 * ߐs(uGridRes+1) * (vGridRes)B
	 * 
	 * @param uGridRes
	 * @param vGridRes
	 */
	public void Initialise(int uGridRes, int vGridRes) {
		CqVector3D Def = new CqVector3D();
		
		if ( m_aValue.size() > 0 )
			Def.assignment(m_aValue.elementAt(0));
		
		if( Def.x == 0.0f && Def.y == 0.0f && Def.z == 0.0f )
		{
			m_aValue.resize( (( uGridRes + 1 ) * ( vGridRes + 1 )) );
		}
		else
		{
			m_aValue.assign( (( uGridRes + 1 ) * ( vGridRes + 1 )), Def );
		}
		
		//܂
		CqStats.STATS_INC( SHD_Var_varying_init );
	}
	
	/**
	 * w肳ꂽꏊShaderData̒l擾
	 * @param res 擾li[ϐ
	 * @param index 擾ꏊ
	 */
	public	void	GetPoint( CqVector3D res, int index)
	{
		res.assignment(m_aValue.elementAt( index ));
	}
	
	/**
	 * ShaderData̎ACqVector3D^̃|C^擾B
	 * 
	 * @param res ʂi[BgύXShaderData̒le󂯂
	 */
	
	public	void	GetPointPtr( STLVector<CqVector3D> res )
	{
		res.resize(m_aValue.size());
		for(int i=0;i < m_aValue.size();i++){
			res.setElementAt(m_aValue.elementAt(i),i);
		}
	}

//	nttdata 
	//public	STLVector<CqVector3D> GetPointPtr_2( )
	public	STLArray<CqVector3D> GetPointPtr_2( )
	{
		return m_aValue;
	}
	/**
	 * ShaderData̎w肳ꂽꏊ̒lύXB
	 * 
	 * @param p ύXl
	 * @param index ύXꏊ
	 */
	public	void	SetPoint( final CqVector3D p, int index )
	{
		m_aValue.elementAt(index).assignment( p );
	}
	
	/**
	 * ShaderData̒lύXB
	 * Ŏw肳ꂽlVector𖄂ߐs
	 * @param p ύXl
	 */
	public	void	SetPoint( final CqVector3D p )
	{
		m_aValue.assign( m_aValue.size(), p );
	}
	
	/**
	 * w肳ꂽꏊShaderData̒l擾
	 * @param res 擾li[ϐ
	 * @param index 擾ꏊ
	 */
	public	void	GetVector( CqVector3D res, int index )
	{
		GetPoint( res, index );
	}
	
	/**
	 * ShaderData̎ACqVector3D^̃|C^擾B
	 * 
	 * @param res ʂi[BgύXShaderData̒le󂯂
	 */
	public	void	GetVectorPtr( STLVector<CqVector3D> res )
	{
		res.resize(m_aValue.size());
		for(int i=0;i < m_aValue.size();i++){
			res.setElementAt(m_aValue.elementAt(i),i);
		}
	}
	//public	STLVector<CqVector3D> GetVectorPtr_2( )
	public	STLArray<CqVector3D> GetVectorPtr_2( )
	{
		return m_aValue;
	}
	/**
	 * ShaderData̎w肳ꂽꏊ̒lύXB
	 * 
	 * @param p ύXl
	 * @param index ύXꏊ
	 */
	public	void	SetVector( final CqVector3D p, int index )
	{
		SetPoint( p, index );
	}
	
	/**
	 * ShaderData̒lύXB
	 * Ŏw肳ꂽlVector𖄂ߐs
	 * @param p ύXl
	 */
	public	void	SetVector( final CqVector3D p )
	{
		SetPoint( p );
	}
	
	/**
	 * w肳ꂽꏊShaderData̒l擾
	 * @param res 擾li[ϐ
	 * @param index 擾ꏊ
	 */
	public	void	GetNormal( CqVector3D res, int index )
	{
		GetPoint( res, index );
	}
	
	/**
	 * ShaderData̎ACqVector3D^̃|C^擾B
	 * 
	 * @param res ʂi[BgύXShaderData̒le󂯂
	 */
	public	void	GetNormalPtr( STLVector<CqVector3D> res )
	{
		res.resize(m_aValue.size());
		for(int i=0;i < m_aValue.size();i++){
			res.setElementAt(m_aValue.elementAt(i),i);
		}
	}
	//public	STLVector<CqVector3D>	GetNormalPtr_2( )
	public	STLArray<CqVector3D>	GetNormalPtr_2( )
	{
		return m_aValue;
	}
	/**
	 * ShaderData̎w肳ꂽꏊ̒lύXB
	 * 
	 * @param p ύXl
	 * @param index ύXꏊ
	 */
	public	void	SetNormal( final CqVector3D p, int index )
	{
		SetPoint( p, index );
	}
	
	/**
	 * ShaderData̒lύXB
	 * Ŏw肳ꂽlVector𖄂ߐs
	 * @param p ύXl
	 */
	public	void	SetNormal( final CqVector3D p )
	{
		SetPoint( p );
	}
	
	/**
	 * w肳ꂽꏊ̒lׂ0̎falseA
	 * ȊOtrueԂB
	 * 
	 * @param res li[ϐ
	 * @param index w肷ꏊ
	 */
	public void	GetBool( p_boolean res, int index )
	{
		res.value = ( ( m_aValue.elementAt( index ).x != 0.0f ) || ( m_aValue.elementAt( index ).y != 0.0f ) || ( m_aValue.elementAt( index ).z != 0.0f ) );
	}
	
	
	/** Copy the values from the passed variable into this, taking into account any class differences.
	 * Ŏw肳ꂽϐlARs[B
	 * Rs[̃TCY1ȉ̏ꍇ́A̒l
	 * Vector̒Rs[B
	 *  
	 * Rs[Vector̃TCY1傫
	 * Vector̃TCY菬ꍇErrorƂȂB
	 * @param pVal The variable to copy from.
	 */
	public	void	SetValueFromVariable( final IqShaderData pVal )
	{
		int i;
		if ( pVal.Size() > 1 )
		{
			//nttdata 
//			STLVector<CqVector3D> pData = new STLVector<CqVector3D>(CqVector3D.class);
//			pVal.GetPointPtr( pData );
			
			//STLVector<CqVector3D> pData;
			STLArray<CqVector3D> pData;
			pData = pVal.GetPointPtr_2( );
			
			for ( i = 0; i < Size(); i++ )
				m_aValue.elementAt( i ).assignment( pData.elementAt( i ));
		}
		else
		{
			CqVector3D temp = new CqVector3D();
			pVal.GetPoint( temp );
			m_aValue.assign( m_aValue.size(), temp );
		}
	}
	
	/** 
	 * Set the all SIMD data ased on a state vector, only indexes whose bit is set are modified.
	 * Ŏw肳ꂽϐVector̎w肳ꂽꏊ̒l
	 * gVector̓ꏊɃRs[B
	 * 
	 * AgA܂̓Rs[̎Vector̒ȏ̏ꏊw肷
	 * G[B(C++ł͓ɃG[NȂ)
	 * @param pVal Rs[ The stack entry to assign.
	 * @param index Rs[ꏊ Integer SIMD index.
	 */
	public	void	SetValueFromVariable( final IqShaderData pVal, int index )
	{
		pVal.GetPoint( m_aValue.elementAt( index ), index );
	}
	
	/**
	 * ƓlCX^XԂ
	 * 
	 * @return g̃Rs[
	 */
	public	IqShaderData Clone()
	{
		CqShaderVariableVaryingPoint newVar = new CqShaderVariableVaryingPoint( this );
		newVar.SetSize( (int) Size() );
		newVar.SetValueFromVariable( this );
		return ( newVar );
	}
	
}
